#define ALGO_CN_0           0x63150000
#define ALGO_CN_1           0x63150100
#define ALGO_CN_2           0x63150200
#define ALGO_CN_R           0x63150272
#define ALGO_CN_FAST        0x63150166
#define ALGO_CN_HALF        0x63150268
#define ALGO_CN_XAO         0x63150078
#define ALGO_CN_RTO         0x63150172
#define ALGO_CN_RWZ         0x63150277
#define ALGO_CN_ZLS         0x6315027a
#define ALGO_CN_DOUBLE      0x63150264
#define ALGO_CN_CCX         0x63150063
#define ALGO_CN_LITE_0      0x63140000
#define ALGO_CN_LITE_1      0x63140100
#define ALGO_CN_HEAVY_0     0x63160000
#define ALGO_CN_HEAVY_TUBE  0x63160172
#define ALGO_CN_HEAVY_XHV   0x63160068
#define ALGO_CN_PICO_0      0x63120200
#define ALGO_CN_PICO_TLO    0x63120274
#define ALGO_CN_UPX2        0x63110200
#define ALGO_RX_0           0x72151200
#define ALGO_RX_WOW         0x72141177
#define ALGO_RX_ARQMA       0x72121061
#define ALGO_RX_SFX         0x72151273
#define ALGO_RX_GRAFT       0x72151267
#define ALGO_RX_YADA        0x72151279
#define ALGO_AR2_CHUKWA     0x61130000
#define ALGO_AR2_CHUKWA_V2  0x61140000
#define ALGO_AR2_WRKZ       0x61120000
#define ALGO_KAWPOW_RVN     0x6b0f0000

#define FAMILY_UNKNOWN      0
#define FAMILY_CN           0x63150000
#define FAMILY_CN_LITE      0x63140000
#define FAMILY_CN_HEAVY     0x63160000
#define FAMILY_CN_PICO      0x63120000
#define FAMILY_CN_FEMTO     0x63110000
#define FAMILY_RANDOM_X     0x72000000
#define FAMILY_ARGON2       0x61000000
#define FAMILY_KAWPOW       0x6b000000
