; Auto-generated file, do not edit

_TEXT_CN_TEMPLATE SEGMENT PAGE READ EXECUTE
PUBLIC CryptonightR_instruction0
PUBLIC CryptonightR_instruction1
PUBLIC CryptonightR_instruction2
PUBLIC CryptonightR_instruction3
PUBLIC CryptonightR_instruction4
PUBLIC CryptonightR_instruction5
PUBLIC CryptonightR_instruction6
PUBLIC CryptonightR_instruction7
PUBLIC CryptonightR_instruction8
PUBLIC CryptonightR_instruction9
PUBLIC CryptonightR_instruction10
PUBLIC CryptonightR_instruction11
PUBLIC CryptonightR_instruction12
PUBLIC CryptonightR_instruction13
PUBLIC CryptonightR_instruction14
PUBLIC CryptonightR_instruction15
PUBLIC CryptonightR_instruction16
PUBLIC CryptonightR_instruction17
PUBLIC CryptonightR_instruction18
PUBLIC CryptonightR_instruction19
PUBLIC CryptonightR_instruction20
PUBLIC CryptonightR_instruction21
PUBLIC CryptonightR_instruction22
PUBLIC CryptonightR_instruction23
PUBLIC CryptonightR_instruction24
PUBLIC CryptonightR_instruction25
PUBLIC CryptonightR_instruction26
PUBLIC CryptonightR_instruction27
PUBLIC CryptonightR_instruction28
PUBLIC CryptonightR_instruction29
PUBLIC CryptonightR_instruction30
PUBLIC CryptonightR_instruction31
PUBLIC CryptonightR_instruction32
PUBLIC CryptonightR_instruction33
PUBLIC CryptonightR_instruction34
PUBLIC CryptonightR_instruction35
PUBLIC CryptonightR_instruction36
PUBLIC CryptonightR_instruction37
PUBLIC CryptonightR_instruction38
PUBLIC CryptonightR_instruction39
PUBLIC CryptonightR_instruction40
PUBLIC CryptonightR_instruction41
PUBLIC CryptonightR_instruction42
PUBLIC CryptonightR_instruction43
PUBLIC CryptonightR_instruction44
PUBLIC CryptonightR_instruction45
PUBLIC CryptonightR_instruction46
PUBLIC CryptonightR_instruction47
PUBLIC CryptonightR_instruction48
PUBLIC CryptonightR_instruction49
PUBLIC CryptonightR_instruction50
PUBLIC CryptonightR_instruction51
PUBLIC CryptonightR_instruction52
PUBLIC CryptonightR_instruction53
PUBLIC CryptonightR_instruction54
PUBLIC CryptonightR_instruction55
PUBLIC CryptonightR_instruction56
PUBLIC CryptonightR_instruction57
PUBLIC CryptonightR_instruction58
PUBLIC CryptonightR_instruction59
PUBLIC CryptonightR_instruction60
PUBLIC CryptonightR_instruction61
PUBLIC CryptonightR_instruction62
PUBLIC CryptonightR_instruction63
PUBLIC CryptonightR_instruction64
PUBLIC CryptonightR_instruction65
PUBLIC CryptonightR_instruction66
PUBLIC CryptonightR_instruction67
PUBLIC CryptonightR_instruction68
PUBLIC CryptonightR_instruction69
PUBLIC CryptonightR_instruction70
PUBLIC CryptonightR_instruction71
PUBLIC CryptonightR_instruction72
PUBLIC CryptonightR_instruction73
PUBLIC CryptonightR_instruction74
PUBLIC CryptonightR_instruction75
PUBLIC CryptonightR_instruction76
PUBLIC CryptonightR_instruction77
PUBLIC CryptonightR_instruction78
PUBLIC CryptonightR_instruction79
PUBLIC CryptonightR_instruction80
PUBLIC CryptonightR_instruction81
PUBLIC CryptonightR_instruction82
PUBLIC CryptonightR_instruction83
PUBLIC CryptonightR_instruction84
PUBLIC CryptonightR_instruction85
PUBLIC CryptonightR_instruction86
PUBLIC CryptonightR_instruction87
PUBLIC CryptonightR_instruction88
PUBLIC CryptonightR_instruction89
PUBLIC CryptonightR_instruction90
PUBLIC CryptonightR_instruction91
PUBLIC CryptonightR_instruction92
PUBLIC CryptonightR_instruction93
PUBLIC CryptonightR_instruction94
PUBLIC CryptonightR_instruction95
PUBLIC CryptonightR_instruction96
PUBLIC CryptonightR_instruction97
PUBLIC CryptonightR_instruction98
PUBLIC CryptonightR_instruction99
PUBLIC CryptonightR_instruction100
PUBLIC CryptonightR_instruction101
PUBLIC CryptonightR_instruction102
PUBLIC CryptonightR_instruction103
PUBLIC CryptonightR_instruction104
PUBLIC CryptonightR_instruction105
PUBLIC CryptonightR_instruction106
PUBLIC CryptonightR_instruction107
PUBLIC CryptonightR_instruction108
PUBLIC CryptonightR_instruction109
PUBLIC CryptonightR_instruction110
PUBLIC CryptonightR_instruction111
PUBLIC CryptonightR_instruction112
PUBLIC CryptonightR_instruction113
PUBLIC CryptonightR_instruction114
PUBLIC CryptonightR_instruction115
PUBLIC CryptonightR_instruction116
PUBLIC CryptonightR_instruction117
PUBLIC CryptonightR_instruction118
PUBLIC CryptonightR_instruction119
PUBLIC CryptonightR_instruction120
PUBLIC CryptonightR_instruction121
PUBLIC CryptonightR_instruction122
PUBLIC CryptonightR_instruction123
PUBLIC CryptonightR_instruction124
PUBLIC CryptonightR_instruction125
PUBLIC CryptonightR_instruction126
PUBLIC CryptonightR_instruction127
PUBLIC CryptonightR_instruction128
PUBLIC CryptonightR_instruction129
PUBLIC CryptonightR_instruction130
PUBLIC CryptonightR_instruction131
PUBLIC CryptonightR_instruction132
PUBLIC CryptonightR_instruction133
PUBLIC CryptonightR_instruction134
PUBLIC CryptonightR_instruction135
PUBLIC CryptonightR_instruction136
PUBLIC CryptonightR_instruction137
PUBLIC CryptonightR_instruction138
PUBLIC CryptonightR_instruction139
PUBLIC CryptonightR_instruction140
PUBLIC CryptonightR_instruction141
PUBLIC CryptonightR_instruction142
PUBLIC CryptonightR_instruction143
PUBLIC CryptonightR_instruction144
PUBLIC CryptonightR_instruction145
PUBLIC CryptonightR_instruction146
PUBLIC CryptonightR_instruction147
PUBLIC CryptonightR_instruction148
PUBLIC CryptonightR_instruction149
PUBLIC CryptonightR_instruction150
PUBLIC CryptonightR_instruction151
PUBLIC CryptonightR_instruction152
PUBLIC CryptonightR_instruction153
PUBLIC CryptonightR_instruction154
PUBLIC CryptonightR_instruction155
PUBLIC CryptonightR_instruction156
PUBLIC CryptonightR_instruction157
PUBLIC CryptonightR_instruction158
PUBLIC CryptonightR_instruction159
PUBLIC CryptonightR_instruction160
PUBLIC CryptonightR_instruction161
PUBLIC CryptonightR_instruction162
PUBLIC CryptonightR_instruction163
PUBLIC CryptonightR_instruction164
PUBLIC CryptonightR_instruction165
PUBLIC CryptonightR_instruction166
PUBLIC CryptonightR_instruction167
PUBLIC CryptonightR_instruction168
PUBLIC CryptonightR_instruction169
PUBLIC CryptonightR_instruction170
PUBLIC CryptonightR_instruction171
PUBLIC CryptonightR_instruction172
PUBLIC CryptonightR_instruction173
PUBLIC CryptonightR_instruction174
PUBLIC CryptonightR_instruction175
PUBLIC CryptonightR_instruction176
PUBLIC CryptonightR_instruction177
PUBLIC CryptonightR_instruction178
PUBLIC CryptonightR_instruction179
PUBLIC CryptonightR_instruction180
PUBLIC CryptonightR_instruction181
PUBLIC CryptonightR_instruction182
PUBLIC CryptonightR_instruction183
PUBLIC CryptonightR_instruction184
PUBLIC CryptonightR_instruction185
PUBLIC CryptonightR_instruction186
PUBLIC CryptonightR_instruction187
PUBLIC CryptonightR_instruction188
PUBLIC CryptonightR_instruction189
PUBLIC CryptonightR_instruction190
PUBLIC CryptonightR_instruction191
PUBLIC CryptonightR_instruction192
PUBLIC CryptonightR_instruction193
PUBLIC CryptonightR_instruction194
PUBLIC CryptonightR_instruction195
PUBLIC CryptonightR_instruction196
PUBLIC CryptonightR_instruction197
PUBLIC CryptonightR_instruction198
PUBLIC CryptonightR_instruction199
PUBLIC CryptonightR_instruction200
PUBLIC CryptonightR_instruction201
PUBLIC CryptonightR_instruction202
PUBLIC CryptonightR_instruction203
PUBLIC CryptonightR_instruction204
PUBLIC CryptonightR_instruction205
PUBLIC CryptonightR_instruction206
PUBLIC CryptonightR_instruction207
PUBLIC CryptonightR_instruction208
PUBLIC CryptonightR_instruction209
PUBLIC CryptonightR_instruction210
PUBLIC CryptonightR_instruction211
PUBLIC CryptonightR_instruction212
PUBLIC CryptonightR_instruction213
PUBLIC CryptonightR_instruction214
PUBLIC CryptonightR_instruction215
PUBLIC CryptonightR_instruction216
PUBLIC CryptonightR_instruction217
PUBLIC CryptonightR_instruction218
PUBLIC CryptonightR_instruction219
PUBLIC CryptonightR_instruction220
PUBLIC CryptonightR_instruction221
PUBLIC CryptonightR_instruction222
PUBLIC CryptonightR_instruction223
PUBLIC CryptonightR_instruction224
PUBLIC CryptonightR_instruction225
PUBLIC CryptonightR_instruction226
PUBLIC CryptonightR_instruction227
PUBLIC CryptonightR_instruction228
PUBLIC CryptonightR_instruction229
PUBLIC CryptonightR_instruction230
PUBLIC CryptonightR_instruction231
PUBLIC CryptonightR_instruction232
PUBLIC CryptonightR_instruction233
PUBLIC CryptonightR_instruction234
PUBLIC CryptonightR_instruction235
PUBLIC CryptonightR_instruction236
PUBLIC CryptonightR_instruction237
PUBLIC CryptonightR_instruction238
PUBLIC CryptonightR_instruction239
PUBLIC CryptonightR_instruction240
PUBLIC CryptonightR_instruction241
PUBLIC CryptonightR_instruction242
PUBLIC CryptonightR_instruction243
PUBLIC CryptonightR_instruction244
PUBLIC CryptonightR_instruction245
PUBLIC CryptonightR_instruction246
PUBLIC CryptonightR_instruction247
PUBLIC CryptonightR_instruction248
PUBLIC CryptonightR_instruction249
PUBLIC CryptonightR_instruction250
PUBLIC CryptonightR_instruction251
PUBLIC CryptonightR_instruction252
PUBLIC CryptonightR_instruction253
PUBLIC CryptonightR_instruction254
PUBLIC CryptonightR_instruction255
PUBLIC CryptonightR_instruction256
PUBLIC CryptonightR_instruction_mov0
PUBLIC CryptonightR_instruction_mov1
PUBLIC CryptonightR_instruction_mov2
PUBLIC CryptonightR_instruction_mov3
PUBLIC CryptonightR_instruction_mov4
PUBLIC CryptonightR_instruction_mov5
PUBLIC CryptonightR_instruction_mov6
PUBLIC CryptonightR_instruction_mov7
PUBLIC CryptonightR_instruction_mov8
PUBLIC CryptonightR_instruction_mov9
PUBLIC CryptonightR_instruction_mov10
PUBLIC CryptonightR_instruction_mov11
PUBLIC CryptonightR_instruction_mov12
PUBLIC CryptonightR_instruction_mov13
PUBLIC CryptonightR_instruction_mov14
PUBLIC CryptonightR_instruction_mov15
PUBLIC CryptonightR_instruction_mov16
PUBLIC CryptonightR_instruction_mov17
PUBLIC CryptonightR_instruction_mov18
PUBLIC CryptonightR_instruction_mov19
PUBLIC CryptonightR_instruction_mov20
PUBLIC CryptonightR_instruction_mov21
PUBLIC CryptonightR_instruction_mov22
PUBLIC CryptonightR_instruction_mov23
PUBLIC CryptonightR_instruction_mov24
PUBLIC CryptonightR_instruction_mov25
PUBLIC CryptonightR_instruction_mov26
PUBLIC CryptonightR_instruction_mov27
PUBLIC CryptonightR_instruction_mov28
PUBLIC CryptonightR_instruction_mov29
PUBLIC CryptonightR_instruction_mov30
PUBLIC CryptonightR_instruction_mov31
PUBLIC CryptonightR_instruction_mov32
PUBLIC CryptonightR_instruction_mov33
PUBLIC CryptonightR_instruction_mov34
PUBLIC CryptonightR_instruction_mov35
PUBLIC CryptonightR_instruction_mov36
PUBLIC CryptonightR_instruction_mov37
PUBLIC CryptonightR_instruction_mov38
PUBLIC CryptonightR_instruction_mov39
PUBLIC CryptonightR_instruction_mov40
PUBLIC CryptonightR_instruction_mov41
PUBLIC CryptonightR_instruction_mov42
PUBLIC CryptonightR_instruction_mov43
PUBLIC CryptonightR_instruction_mov44
PUBLIC CryptonightR_instruction_mov45
PUBLIC CryptonightR_instruction_mov46
PUBLIC CryptonightR_instruction_mov47
PUBLIC CryptonightR_instruction_mov48
PUBLIC CryptonightR_instruction_mov49
PUBLIC CryptonightR_instruction_mov50
PUBLIC CryptonightR_instruction_mov51
PUBLIC CryptonightR_instruction_mov52
PUBLIC CryptonightR_instruction_mov53
PUBLIC CryptonightR_instruction_mov54
PUBLIC CryptonightR_instruction_mov55
PUBLIC CryptonightR_instruction_mov56
PUBLIC CryptonightR_instruction_mov57
PUBLIC CryptonightR_instruction_mov58
PUBLIC CryptonightR_instruction_mov59
PUBLIC CryptonightR_instruction_mov60
PUBLIC CryptonightR_instruction_mov61
PUBLIC CryptonightR_instruction_mov62
PUBLIC CryptonightR_instruction_mov63
PUBLIC CryptonightR_instruction_mov64
PUBLIC CryptonightR_instruction_mov65
PUBLIC CryptonightR_instruction_mov66
PUBLIC CryptonightR_instruction_mov67
PUBLIC CryptonightR_instruction_mov68
PUBLIC CryptonightR_instruction_mov69
PUBLIC CryptonightR_instruction_mov70
PUBLIC CryptonightR_instruction_mov71
PUBLIC CryptonightR_instruction_mov72
PUBLIC CryptonightR_instruction_mov73
PUBLIC CryptonightR_instruction_mov74
PUBLIC CryptonightR_instruction_mov75
PUBLIC CryptonightR_instruction_mov76
PUBLIC CryptonightR_instruction_mov77
PUBLIC CryptonightR_instruction_mov78
PUBLIC CryptonightR_instruction_mov79
PUBLIC CryptonightR_instruction_mov80
PUBLIC CryptonightR_instruction_mov81
PUBLIC CryptonightR_instruction_mov82
PUBLIC CryptonightR_instruction_mov83
PUBLIC CryptonightR_instruction_mov84
PUBLIC CryptonightR_instruction_mov85
PUBLIC CryptonightR_instruction_mov86
PUBLIC CryptonightR_instruction_mov87
PUBLIC CryptonightR_instruction_mov88
PUBLIC CryptonightR_instruction_mov89
PUBLIC CryptonightR_instruction_mov90
PUBLIC CryptonightR_instruction_mov91
PUBLIC CryptonightR_instruction_mov92
PUBLIC CryptonightR_instruction_mov93
PUBLIC CryptonightR_instruction_mov94
PUBLIC CryptonightR_instruction_mov95
PUBLIC CryptonightR_instruction_mov96
PUBLIC CryptonightR_instruction_mov97
PUBLIC CryptonightR_instruction_mov98
PUBLIC CryptonightR_instruction_mov99
PUBLIC CryptonightR_instruction_mov100
PUBLIC CryptonightR_instruction_mov101
PUBLIC CryptonightR_instruction_mov102
PUBLIC CryptonightR_instruction_mov103
PUBLIC CryptonightR_instruction_mov104
PUBLIC CryptonightR_instruction_mov105
PUBLIC CryptonightR_instruction_mov106
PUBLIC CryptonightR_instruction_mov107
PUBLIC CryptonightR_instruction_mov108
PUBLIC CryptonightR_instruction_mov109
PUBLIC CryptonightR_instruction_mov110
PUBLIC CryptonightR_instruction_mov111
PUBLIC CryptonightR_instruction_mov112
PUBLIC CryptonightR_instruction_mov113
PUBLIC CryptonightR_instruction_mov114
PUBLIC CryptonightR_instruction_mov115
PUBLIC CryptonightR_instruction_mov116
PUBLIC CryptonightR_instruction_mov117
PUBLIC CryptonightR_instruction_mov118
PUBLIC CryptonightR_instruction_mov119
PUBLIC CryptonightR_instruction_mov120
PUBLIC CryptonightR_instruction_mov121
PUBLIC CryptonightR_instruction_mov122
PUBLIC CryptonightR_instruction_mov123
PUBLIC CryptonightR_instruction_mov124
PUBLIC CryptonightR_instruction_mov125
PUBLIC CryptonightR_instruction_mov126
PUBLIC CryptonightR_instruction_mov127
PUBLIC CryptonightR_instruction_mov128
PUBLIC CryptonightR_instruction_mov129
PUBLIC CryptonightR_instruction_mov130
PUBLIC CryptonightR_instruction_mov131
PUBLIC CryptonightR_instruction_mov132
PUBLIC CryptonightR_instruction_mov133
PUBLIC CryptonightR_instruction_mov134
PUBLIC CryptonightR_instruction_mov135
PUBLIC CryptonightR_instruction_mov136
PUBLIC CryptonightR_instruction_mov137
PUBLIC CryptonightR_instruction_mov138
PUBLIC CryptonightR_instruction_mov139
PUBLIC CryptonightR_instruction_mov140
PUBLIC CryptonightR_instruction_mov141
PUBLIC CryptonightR_instruction_mov142
PUBLIC CryptonightR_instruction_mov143
PUBLIC CryptonightR_instruction_mov144
PUBLIC CryptonightR_instruction_mov145
PUBLIC CryptonightR_instruction_mov146
PUBLIC CryptonightR_instruction_mov147
PUBLIC CryptonightR_instruction_mov148
PUBLIC CryptonightR_instruction_mov149
PUBLIC CryptonightR_instruction_mov150
PUBLIC CryptonightR_instruction_mov151
PUBLIC CryptonightR_instruction_mov152
PUBLIC CryptonightR_instruction_mov153
PUBLIC CryptonightR_instruction_mov154
PUBLIC CryptonightR_instruction_mov155
PUBLIC CryptonightR_instruction_mov156
PUBLIC CryptonightR_instruction_mov157
PUBLIC CryptonightR_instruction_mov158
PUBLIC CryptonightR_instruction_mov159
PUBLIC CryptonightR_instruction_mov160
PUBLIC CryptonightR_instruction_mov161
PUBLIC CryptonightR_instruction_mov162
PUBLIC CryptonightR_instruction_mov163
PUBLIC CryptonightR_instruction_mov164
PUBLIC CryptonightR_instruction_mov165
PUBLIC CryptonightR_instruction_mov166
PUBLIC CryptonightR_instruction_mov167
PUBLIC CryptonightR_instruction_mov168
PUBLIC CryptonightR_instruction_mov169
PUBLIC CryptonightR_instruction_mov170
PUBLIC CryptonightR_instruction_mov171
PUBLIC CryptonightR_instruction_mov172
PUBLIC CryptonightR_instruction_mov173
PUBLIC CryptonightR_instruction_mov174
PUBLIC CryptonightR_instruction_mov175
PUBLIC CryptonightR_instruction_mov176
PUBLIC CryptonightR_instruction_mov177
PUBLIC CryptonightR_instruction_mov178
PUBLIC CryptonightR_instruction_mov179
PUBLIC CryptonightR_instruction_mov180
PUBLIC CryptonightR_instruction_mov181
PUBLIC CryptonightR_instruction_mov182
PUBLIC CryptonightR_instruction_mov183
PUBLIC CryptonightR_instruction_mov184
PUBLIC CryptonightR_instruction_mov185
PUBLIC CryptonightR_instruction_mov186
PUBLIC CryptonightR_instruction_mov187
PUBLIC CryptonightR_instruction_mov188
PUBLIC CryptonightR_instruction_mov189
PUBLIC CryptonightR_instruction_mov190
PUBLIC CryptonightR_instruction_mov191
PUBLIC CryptonightR_instruction_mov192
PUBLIC CryptonightR_instruction_mov193
PUBLIC CryptonightR_instruction_mov194
PUBLIC CryptonightR_instruction_mov195
PUBLIC CryptonightR_instruction_mov196
PUBLIC CryptonightR_instruction_mov197
PUBLIC CryptonightR_instruction_mov198
PUBLIC CryptonightR_instruction_mov199
PUBLIC CryptonightR_instruction_mov200
PUBLIC CryptonightR_instruction_mov201
PUBLIC CryptonightR_instruction_mov202
PUBLIC CryptonightR_instruction_mov203
PUBLIC CryptonightR_instruction_mov204
PUBLIC CryptonightR_instruction_mov205
PUBLIC CryptonightR_instruction_mov206
PUBLIC CryptonightR_instruction_mov207
PUBLIC CryptonightR_instruction_mov208
PUBLIC CryptonightR_instruction_mov209
PUBLIC CryptonightR_instruction_mov210
PUBLIC CryptonightR_instruction_mov211
PUBLIC CryptonightR_instruction_mov212
PUBLIC CryptonightR_instruction_mov213
PUBLIC CryptonightR_instruction_mov214
PUBLIC CryptonightR_instruction_mov215
PUBLIC CryptonightR_instruction_mov216
PUBLIC CryptonightR_instruction_mov217
PUBLIC CryptonightR_instruction_mov218
PUBLIC CryptonightR_instruction_mov219
PUBLIC CryptonightR_instruction_mov220
PUBLIC CryptonightR_instruction_mov221
PUBLIC CryptonightR_instruction_mov222
PUBLIC CryptonightR_instruction_mov223
PUBLIC CryptonightR_instruction_mov224
PUBLIC CryptonightR_instruction_mov225
PUBLIC CryptonightR_instruction_mov226
PUBLIC CryptonightR_instruction_mov227
PUBLIC CryptonightR_instruction_mov228
PUBLIC CryptonightR_instruction_mov229
PUBLIC CryptonightR_instruction_mov230
PUBLIC CryptonightR_instruction_mov231
PUBLIC CryptonightR_instruction_mov232
PUBLIC CryptonightR_instruction_mov233
PUBLIC CryptonightR_instruction_mov234
PUBLIC CryptonightR_instruction_mov235
PUBLIC CryptonightR_instruction_mov236
PUBLIC CryptonightR_instruction_mov237
PUBLIC CryptonightR_instruction_mov238
PUBLIC CryptonightR_instruction_mov239
PUBLIC CryptonightR_instruction_mov240
PUBLIC CryptonightR_instruction_mov241
PUBLIC CryptonightR_instruction_mov242
PUBLIC CryptonightR_instruction_mov243
PUBLIC CryptonightR_instruction_mov244
PUBLIC CryptonightR_instruction_mov245
PUBLIC CryptonightR_instruction_mov246
PUBLIC CryptonightR_instruction_mov247
PUBLIC CryptonightR_instruction_mov248
PUBLIC CryptonightR_instruction_mov249
PUBLIC CryptonightR_instruction_mov250
PUBLIC CryptonightR_instruction_mov251
PUBLIC CryptonightR_instruction_mov252
PUBLIC CryptonightR_instruction_mov253
PUBLIC CryptonightR_instruction_mov254
PUBLIC CryptonightR_instruction_mov255
PUBLIC CryptonightR_instruction_mov256

INCLUDE CryptonightR_template_win.inc
INCLUDE CryptonightR_soft_aes_template_win.inc

CryptonightR_instruction0:
	imul	rbx, rbx
CryptonightR_instruction1:
	imul	rbx, rbx
CryptonightR_instruction2:
	imul	rbx, rbx
CryptonightR_instruction3:
	add	rbx, r9
	add	rbx, 2147483647
CryptonightR_instruction4:
	sub	rbx, r9
CryptonightR_instruction5:
	ror	ebx, cl
CryptonightR_instruction6:
	rol	ebx, cl
CryptonightR_instruction7:
	xor	rbx, r9
CryptonightR_instruction8:
	imul	rsi, rbx
CryptonightR_instruction9:
	imul	rsi, rbx
CryptonightR_instruction10:
	imul	rsi, rbx
CryptonightR_instruction11:
	add	rsi, rbx
	add	rsi, 2147483647
CryptonightR_instruction12:
	sub	rsi, rbx
CryptonightR_instruction13:
	ror	esi, cl
CryptonightR_instruction14:
	rol	esi, cl
CryptonightR_instruction15:
	xor	rsi, rbx
CryptonightR_instruction16:
	imul	rdi, rbx
CryptonightR_instruction17:
	imul	rdi, rbx
CryptonightR_instruction18:
	imul	rdi, rbx
CryptonightR_instruction19:
	add	rdi, rbx
	add	rdi, 2147483647
CryptonightR_instruction20:
	sub	rdi, rbx
CryptonightR_instruction21:
	ror	edi, cl
CryptonightR_instruction22:
	rol	edi, cl
CryptonightR_instruction23:
	xor	rdi, rbx
CryptonightR_instruction24:
	imul	rbp, rbx
CryptonightR_instruction25:
	imul	rbp, rbx
CryptonightR_instruction26:
	imul	rbp, rbx
CryptonightR_instruction27:
	add	rbp, rbx
	add	rbp, 2147483647
CryptonightR_instruction28:
	sub	rbp, rbx
CryptonightR_instruction29:
	ror	ebp, cl
CryptonightR_instruction30:
	rol	ebp, cl
CryptonightR_instruction31:
	xor	rbp, rbx
CryptonightR_instruction32:
	imul	rbx, rsi
CryptonightR_instruction33:
	imul	rbx, rsi
CryptonightR_instruction34:
	imul	rbx, rsi
CryptonightR_instruction35:
	add	rbx, rsi
	add	rbx, 2147483647
CryptonightR_instruction36:
	sub	rbx, rsi
CryptonightR_instruction37:
	ror	ebx, cl
CryptonightR_instruction38:
	rol	ebx, cl
CryptonightR_instruction39:
	xor	rbx, rsi
CryptonightR_instruction40:
	imul	rsi, rsi
CryptonightR_instruction41:
	imul	rsi, rsi
CryptonightR_instruction42:
	imul	rsi, rsi
CryptonightR_instruction43:
	add	rsi, r9
	add	rsi, 2147483647
CryptonightR_instruction44:
	sub	rsi, r9
CryptonightR_instruction45:
	ror	esi, cl
CryptonightR_instruction46:
	rol	esi, cl
CryptonightR_instruction47:
	xor	rsi, r9
CryptonightR_instruction48:
	imul	rdi, rsi
CryptonightR_instruction49:
	imul	rdi, rsi
CryptonightR_instruction50:
	imul	rdi, rsi
CryptonightR_instruction51:
	add	rdi, rsi
	add	rdi, 2147483647
CryptonightR_instruction52:
	sub	rdi, rsi
CryptonightR_instruction53:
	ror	edi, cl
CryptonightR_instruction54:
	rol	edi, cl
CryptonightR_instruction55:
	xor	rdi, rsi
CryptonightR_instruction56:
	imul	rbp, rsi
CryptonightR_instruction57:
	imul	rbp, rsi
CryptonightR_instruction58:
	imul	rbp, rsi
CryptonightR_instruction59:
	add	rbp, rsi
	add	rbp, 2147483647
CryptonightR_instruction60:
	sub	rbp, rsi
CryptonightR_instruction61:
	ror	ebp, cl
CryptonightR_instruction62:
	rol	ebp, cl
CryptonightR_instruction63:
	xor	rbp, rsi
CryptonightR_instruction64:
	imul	rbx, rdi
CryptonightR_instruction65:
	imul	rbx, rdi
CryptonightR_instruction66:
	imul	rbx, rdi
CryptonightR_instruction67:
	add	rbx, rdi
	add	rbx, 2147483647
CryptonightR_instruction68:
	sub	rbx, rdi
CryptonightR_instruction69:
	ror	ebx, cl
CryptonightR_instruction70:
	rol	ebx, cl
CryptonightR_instruction71:
	xor	rbx, rdi
CryptonightR_instruction72:
	imul	rsi, rdi
CryptonightR_instruction73:
	imul	rsi, rdi
CryptonightR_instruction74:
	imul	rsi, rdi
CryptonightR_instruction75:
	add	rsi, rdi
	add	rsi, 2147483647
CryptonightR_instruction76:
	sub	rsi, rdi
CryptonightR_instruction77:
	ror	esi, cl
CryptonightR_instruction78:
	rol	esi, cl
CryptonightR_instruction79:
	xor	rsi, rdi
CryptonightR_instruction80:
	imul	rdi, rdi
CryptonightR_instruction81:
	imul	rdi, rdi
CryptonightR_instruction82:
	imul	rdi, rdi
CryptonightR_instruction83:
	add	rdi, r9
	add	rdi, 2147483647
CryptonightR_instruction84:
	sub	rdi, r9
CryptonightR_instruction85:
	ror	edi, cl
CryptonightR_instruction86:
	rol	edi, cl
CryptonightR_instruction87:
	xor	rdi, r9
CryptonightR_instruction88:
	imul	rbp, rdi
CryptonightR_instruction89:
	imul	rbp, rdi
CryptonightR_instruction90:
	imul	rbp, rdi
CryptonightR_instruction91:
	add	rbp, rdi
	add	rbp, 2147483647
CryptonightR_instruction92:
	sub	rbp, rdi
CryptonightR_instruction93:
	ror	ebp, cl
CryptonightR_instruction94:
	rol	ebp, cl
CryptonightR_instruction95:
	xor	rbp, rdi
CryptonightR_instruction96:
	imul	rbx, rbp
CryptonightR_instruction97:
	imul	rbx, rbp
CryptonightR_instruction98:
	imul	rbx, rbp
CryptonightR_instruction99:
	add	rbx, rbp
	add	rbx, 2147483647
CryptonightR_instruction100:
	sub	rbx, rbp
CryptonightR_instruction101:
	ror	ebx, cl
CryptonightR_instruction102:
	rol	ebx, cl
CryptonightR_instruction103:
	xor	rbx, rbp
CryptonightR_instruction104:
	imul	rsi, rbp
CryptonightR_instruction105:
	imul	rsi, rbp
CryptonightR_instruction106:
	imul	rsi, rbp
CryptonightR_instruction107:
	add	rsi, rbp
	add	rsi, 2147483647
CryptonightR_instruction108:
	sub	rsi, rbp
CryptonightR_instruction109:
	ror	esi, cl
CryptonightR_instruction110:
	rol	esi, cl
CryptonightR_instruction111:
	xor	rsi, rbp
CryptonightR_instruction112:
	imul	rdi, rbp
CryptonightR_instruction113:
	imul	rdi, rbp
CryptonightR_instruction114:
	imul	rdi, rbp
CryptonightR_instruction115:
	add	rdi, rbp
	add	rdi, 2147483647
CryptonightR_instruction116:
	sub	rdi, rbp
CryptonightR_instruction117:
	ror	edi, cl
CryptonightR_instruction118:
	rol	edi, cl
CryptonightR_instruction119:
	xor	rdi, rbp
CryptonightR_instruction120:
	imul	rbp, rbp
CryptonightR_instruction121:
	imul	rbp, rbp
CryptonightR_instruction122:
	imul	rbp, rbp
CryptonightR_instruction123:
	add	rbp, r9
	add	rbp, 2147483647
CryptonightR_instruction124:
	sub	rbp, r9
CryptonightR_instruction125:
	ror	ebp, cl
CryptonightR_instruction126:
	rol	ebp, cl
CryptonightR_instruction127:
	xor	rbp, r9
CryptonightR_instruction128:
	imul	rbx, rsp
CryptonightR_instruction129:
	imul	rbx, rsp
CryptonightR_instruction130:
	imul	rbx, rsp
CryptonightR_instruction131:
	add	rbx, rsp
	add	rbx, 2147483647
CryptonightR_instruction132:
	sub	rbx, rsp
CryptonightR_instruction133:
	ror	ebx, cl
CryptonightR_instruction134:
	rol	ebx, cl
CryptonightR_instruction135:
	xor	rbx, rsp
CryptonightR_instruction136:
	imul	rsi, rsp
CryptonightR_instruction137:
	imul	rsi, rsp
CryptonightR_instruction138:
	imul	rsi, rsp
CryptonightR_instruction139:
	add	rsi, rsp
	add	rsi, 2147483647
CryptonightR_instruction140:
	sub	rsi, rsp
CryptonightR_instruction141:
	ror	esi, cl
CryptonightR_instruction142:
	rol	esi, cl
CryptonightR_instruction143:
	xor	rsi, rsp
CryptonightR_instruction144:
	imul	rdi, rsp
CryptonightR_instruction145:
	imul	rdi, rsp
CryptonightR_instruction146:
	imul	rdi, rsp
CryptonightR_instruction147:
	add	rdi, rsp
	add	rdi, 2147483647
CryptonightR_instruction148:
	sub	rdi, rsp
CryptonightR_instruction149:
	ror	edi, cl
CryptonightR_instruction150:
	rol	edi, cl
CryptonightR_instruction151:
	xor	rdi, rsp
CryptonightR_instruction152:
	imul	rbp, rsp
CryptonightR_instruction153:
	imul	rbp, rsp
CryptonightR_instruction154:
	imul	rbp, rsp
CryptonightR_instruction155:
	add	rbp, rsp
	add	rbp, 2147483647
CryptonightR_instruction156:
	sub	rbp, rsp
CryptonightR_instruction157:
	ror	ebp, cl
CryptonightR_instruction158:
	rol	ebp, cl
CryptonightR_instruction159:
	xor	rbp, rsp
CryptonightR_instruction160:
	imul	rbx, r15
CryptonightR_instruction161:
	imul	rbx, r15
CryptonightR_instruction162:
	imul	rbx, r15
CryptonightR_instruction163:
	add	rbx, r15
	add	rbx, 2147483647
CryptonightR_instruction164:
	sub	rbx, r15
CryptonightR_instruction165:
	ror	ebx, cl
CryptonightR_instruction166:
	rol	ebx, cl
CryptonightR_instruction167:
	xor	rbx, r15
CryptonightR_instruction168:
	imul	rsi, r15
CryptonightR_instruction169:
	imul	rsi, r15
CryptonightR_instruction170:
	imul	rsi, r15
CryptonightR_instruction171:
	add	rsi, r15
	add	rsi, 2147483647
CryptonightR_instruction172:
	sub	rsi, r15
CryptonightR_instruction173:
	ror	esi, cl
CryptonightR_instruction174:
	rol	esi, cl
CryptonightR_instruction175:
	xor	rsi, r15
CryptonightR_instruction176:
	imul	rdi, r15
CryptonightR_instruction177:
	imul	rdi, r15
CryptonightR_instruction178:
	imul	rdi, r15
CryptonightR_instruction179:
	add	rdi, r15
	add	rdi, 2147483647
CryptonightR_instruction180:
	sub	rdi, r15
CryptonightR_instruction181:
	ror	edi, cl
CryptonightR_instruction182:
	rol	edi, cl
CryptonightR_instruction183:
	xor	rdi, r15
CryptonightR_instruction184:
	imul	rbp, r15
CryptonightR_instruction185:
	imul	rbp, r15
CryptonightR_instruction186:
	imul	rbp, r15
CryptonightR_instruction187:
	add	rbp, r15
	add	rbp, 2147483647
CryptonightR_instruction188:
	sub	rbp, r15
CryptonightR_instruction189:
	ror	ebp, cl
CryptonightR_instruction190:
	rol	ebp, cl
CryptonightR_instruction191:
	xor	rbp, r15
CryptonightR_instruction192:
	imul	rbx, rax
CryptonightR_instruction193:
	imul	rbx, rax
CryptonightR_instruction194:
	imul	rbx, rax
CryptonightR_instruction195:
	add	rbx, rax
	add	rbx, 2147483647
CryptonightR_instruction196:
	sub	rbx, rax
CryptonightR_instruction197:
	ror	ebx, cl
CryptonightR_instruction198:
	rol	ebx, cl
CryptonightR_instruction199:
	xor	rbx, rax
CryptonightR_instruction200:
	imul	rsi, rax
CryptonightR_instruction201:
	imul	rsi, rax
CryptonightR_instruction202:
	imul	rsi, rax
CryptonightR_instruction203:
	add	rsi, rax
	add	rsi, 2147483647
CryptonightR_instruction204:
	sub	rsi, rax
CryptonightR_instruction205:
	ror	esi, cl
CryptonightR_instruction206:
	rol	esi, cl
CryptonightR_instruction207:
	xor	rsi, rax
CryptonightR_instruction208:
	imul	rdi, rax
CryptonightR_instruction209:
	imul	rdi, rax
CryptonightR_instruction210:
	imul	rdi, rax
CryptonightR_instruction211:
	add	rdi, rax
	add	rdi, 2147483647
CryptonightR_instruction212:
	sub	rdi, rax
CryptonightR_instruction213:
	ror	edi, cl
CryptonightR_instruction214:
	rol	edi, cl
CryptonightR_instruction215:
	xor	rdi, rax
CryptonightR_instruction216:
	imul	rbp, rax
CryptonightR_instruction217:
	imul	rbp, rax
CryptonightR_instruction218:
	imul	rbp, rax
CryptonightR_instruction219:
	add	rbp, rax
	add	rbp, 2147483647
CryptonightR_instruction220:
	sub	rbp, rax
CryptonightR_instruction221:
	ror	ebp, cl
CryptonightR_instruction222:
	rol	ebp, cl
CryptonightR_instruction223:
	xor	rbp, rax
CryptonightR_instruction224:
	imul	rbx, rdx
CryptonightR_instruction225:
	imul	rbx, rdx
CryptonightR_instruction226:
	imul	rbx, rdx
CryptonightR_instruction227:
	add	rbx, rdx
	add	rbx, 2147483647
CryptonightR_instruction228:
	sub	rbx, rdx
CryptonightR_instruction229:
	ror	ebx, cl
CryptonightR_instruction230:
	rol	ebx, cl
CryptonightR_instruction231:
	xor	rbx, rdx
CryptonightR_instruction232:
	imul	rsi, rdx
CryptonightR_instruction233:
	imul	rsi, rdx
CryptonightR_instruction234:
	imul	rsi, rdx
CryptonightR_instruction235:
	add	rsi, rdx
	add	rsi, 2147483647
CryptonightR_instruction236:
	sub	rsi, rdx
CryptonightR_instruction237:
	ror	esi, cl
CryptonightR_instruction238:
	rol	esi, cl
CryptonightR_instruction239:
	xor	rsi, rdx
CryptonightR_instruction240:
	imul	rdi, rdx
CryptonightR_instruction241:
	imul	rdi, rdx
CryptonightR_instruction242:
	imul	rdi, rdx
CryptonightR_instruction243:
	add	rdi, rdx
	add	rdi, 2147483647
CryptonightR_instruction244:
	sub	rdi, rdx
CryptonightR_instruction245:
	ror	edi, cl
CryptonightR_instruction246:
	rol	edi, cl
CryptonightR_instruction247:
	xor	rdi, rdx
CryptonightR_instruction248:
	imul	rbp, rdx
CryptonightR_instruction249:
	imul	rbp, rdx
CryptonightR_instruction250:
	imul	rbp, rdx
CryptonightR_instruction251:
	add	rbp, rdx
	add	rbp, 2147483647
CryptonightR_instruction252:
	sub	rbp, rdx
CryptonightR_instruction253:
	ror	ebp, cl
CryptonightR_instruction254:
	rol	ebp, cl
CryptonightR_instruction255:
	xor	rbp, rdx
CryptonightR_instruction256:
	imul	rbx, rbx
CryptonightR_instruction_mov0:

CryptonightR_instruction_mov1:

CryptonightR_instruction_mov2:

CryptonightR_instruction_mov3:

CryptonightR_instruction_mov4:

CryptonightR_instruction_mov5:
	mov	rcx, rbx
CryptonightR_instruction_mov6:
	mov	rcx, rbx
CryptonightR_instruction_mov7:

CryptonightR_instruction_mov8:

CryptonightR_instruction_mov9:

CryptonightR_instruction_mov10:

CryptonightR_instruction_mov11:

CryptonightR_instruction_mov12:

CryptonightR_instruction_mov13:
	mov	rcx, rbx
CryptonightR_instruction_mov14:
	mov	rcx, rbx
CryptonightR_instruction_mov15:

CryptonightR_instruction_mov16:

CryptonightR_instruction_mov17:

CryptonightR_instruction_mov18:

CryptonightR_instruction_mov19:

CryptonightR_instruction_mov20:

CryptonightR_instruction_mov21:
	mov	rcx, rbx
CryptonightR_instruction_mov22:
	mov	rcx, rbx
CryptonightR_instruction_mov23:

CryptonightR_instruction_mov24:

CryptonightR_instruction_mov25:

CryptonightR_instruction_mov26:

CryptonightR_instruction_mov27:

CryptonightR_instruction_mov28:

CryptonightR_instruction_mov29:
	mov	rcx, rbx
CryptonightR_instruction_mov30:
	mov	rcx, rbx
CryptonightR_instruction_mov31:

CryptonightR_instruction_mov32:

CryptonightR_instruction_mov33:

CryptonightR_instruction_mov34:

CryptonightR_instruction_mov35:

CryptonightR_instruction_mov36:

CryptonightR_instruction_mov37:
	mov	rcx, rsi
CryptonightR_instruction_mov38:
	mov	rcx, rsi
CryptonightR_instruction_mov39:

CryptonightR_instruction_mov40:

CryptonightR_instruction_mov41:

CryptonightR_instruction_mov42:

CryptonightR_instruction_mov43:

CryptonightR_instruction_mov44:

CryptonightR_instruction_mov45:
	mov	rcx, rsi
CryptonightR_instruction_mov46:
	mov	rcx, rsi
CryptonightR_instruction_mov47:

CryptonightR_instruction_mov48:

CryptonightR_instruction_mov49:

CryptonightR_instruction_mov50:

CryptonightR_instruction_mov51:

CryptonightR_instruction_mov52:

CryptonightR_instruction_mov53:
	mov	rcx, rsi
CryptonightR_instruction_mov54:
	mov	rcx, rsi
CryptonightR_instruction_mov55:

CryptonightR_instruction_mov56:

CryptonightR_instruction_mov57:

CryptonightR_instruction_mov58:

CryptonightR_instruction_mov59:

CryptonightR_instruction_mov60:

CryptonightR_instruction_mov61:
	mov	rcx, rsi
CryptonightR_instruction_mov62:
	mov	rcx, rsi
CryptonightR_instruction_mov63:

CryptonightR_instruction_mov64:

CryptonightR_instruction_mov65:

CryptonightR_instruction_mov66:

CryptonightR_instruction_mov67:

CryptonightR_instruction_mov68:

CryptonightR_instruction_mov69:
	mov	rcx, rdi
CryptonightR_instruction_mov70:
	mov	rcx, rdi
CryptonightR_instruction_mov71:

CryptonightR_instruction_mov72:

CryptonightR_instruction_mov73:

CryptonightR_instruction_mov74:

CryptonightR_instruction_mov75:

CryptonightR_instruction_mov76:

CryptonightR_instruction_mov77:
	mov	rcx, rdi
CryptonightR_instruction_mov78:
	mov	rcx, rdi
CryptonightR_instruction_mov79:

CryptonightR_instruction_mov80:

CryptonightR_instruction_mov81:

CryptonightR_instruction_mov82:

CryptonightR_instruction_mov83:

CryptonightR_instruction_mov84:

CryptonightR_instruction_mov85:
	mov	rcx, rdi
CryptonightR_instruction_mov86:
	mov	rcx, rdi
CryptonightR_instruction_mov87:

CryptonightR_instruction_mov88:

CryptonightR_instruction_mov89:

CryptonightR_instruction_mov90:

CryptonightR_instruction_mov91:

CryptonightR_instruction_mov92:

CryptonightR_instruction_mov93:
	mov	rcx, rdi
CryptonightR_instruction_mov94:
	mov	rcx, rdi
CryptonightR_instruction_mov95:

CryptonightR_instruction_mov96:

CryptonightR_instruction_mov97:

CryptonightR_instruction_mov98:

CryptonightR_instruction_mov99:

CryptonightR_instruction_mov100:

CryptonightR_instruction_mov101:
	mov	rcx, rbp
CryptonightR_instruction_mov102:
	mov	rcx, rbp
CryptonightR_instruction_mov103:

CryptonightR_instruction_mov104:

CryptonightR_instruction_mov105:

CryptonightR_instruction_mov106:

CryptonightR_instruction_mov107:

CryptonightR_instruction_mov108:

CryptonightR_instruction_mov109:
	mov	rcx, rbp
CryptonightR_instruction_mov110:
	mov	rcx, rbp
CryptonightR_instruction_mov111:

CryptonightR_instruction_mov112:

CryptonightR_instruction_mov113:

CryptonightR_instruction_mov114:

CryptonightR_instruction_mov115:

CryptonightR_instruction_mov116:

CryptonightR_instruction_mov117:
	mov	rcx, rbp
CryptonightR_instruction_mov118:
	mov	rcx, rbp
CryptonightR_instruction_mov119:

CryptonightR_instruction_mov120:

CryptonightR_instruction_mov121:

CryptonightR_instruction_mov122:

CryptonightR_instruction_mov123:

CryptonightR_instruction_mov124:

CryptonightR_instruction_mov125:
	mov	rcx, rbp
CryptonightR_instruction_mov126:
	mov	rcx, rbp
CryptonightR_instruction_mov127:

CryptonightR_instruction_mov128:

CryptonightR_instruction_mov129:

CryptonightR_instruction_mov130:

CryptonightR_instruction_mov131:

CryptonightR_instruction_mov132:

CryptonightR_instruction_mov133:
	mov	rcx, rsp
CryptonightR_instruction_mov134:
	mov	rcx, rsp
CryptonightR_instruction_mov135:

CryptonightR_instruction_mov136:

CryptonightR_instruction_mov137:

CryptonightR_instruction_mov138:

CryptonightR_instruction_mov139:

CryptonightR_instruction_mov140:

CryptonightR_instruction_mov141:
	mov	rcx, rsp
CryptonightR_instruction_mov142:
	mov	rcx, rsp
CryptonightR_instruction_mov143:

CryptonightR_instruction_mov144:

CryptonightR_instruction_mov145:

CryptonightR_instruction_mov146:

CryptonightR_instruction_mov147:

CryptonightR_instruction_mov148:

CryptonightR_instruction_mov149:
	mov	rcx, rsp
CryptonightR_instruction_mov150:
	mov	rcx, rsp
CryptonightR_instruction_mov151:

CryptonightR_instruction_mov152:

CryptonightR_instruction_mov153:

CryptonightR_instruction_mov154:

CryptonightR_instruction_mov155:

CryptonightR_instruction_mov156:

CryptonightR_instruction_mov157:
	mov	rcx, rsp
CryptonightR_instruction_mov158:
	mov	rcx, rsp
CryptonightR_instruction_mov159:

CryptonightR_instruction_mov160:

CryptonightR_instruction_mov161:

CryptonightR_instruction_mov162:

CryptonightR_instruction_mov163:

CryptonightR_instruction_mov164:

CryptonightR_instruction_mov165:
	mov	rcx, r15
CryptonightR_instruction_mov166:
	mov	rcx, r15
CryptonightR_instruction_mov167:

CryptonightR_instruction_mov168:

CryptonightR_instruction_mov169:

CryptonightR_instruction_mov170:

CryptonightR_instruction_mov171:

CryptonightR_instruction_mov172:

CryptonightR_instruction_mov173:
	mov	rcx, r15
CryptonightR_instruction_mov174:
	mov	rcx, r15
CryptonightR_instruction_mov175:

CryptonightR_instruction_mov176:

CryptonightR_instruction_mov177:

CryptonightR_instruction_mov178:

CryptonightR_instruction_mov179:

CryptonightR_instruction_mov180:

CryptonightR_instruction_mov181:
	mov	rcx, r15
CryptonightR_instruction_mov182:
	mov	rcx, r15
CryptonightR_instruction_mov183:

CryptonightR_instruction_mov184:

CryptonightR_instruction_mov185:

CryptonightR_instruction_mov186:

CryptonightR_instruction_mov187:

CryptonightR_instruction_mov188:

CryptonightR_instruction_mov189:
	mov	rcx, r15
CryptonightR_instruction_mov190:
	mov	rcx, r15
CryptonightR_instruction_mov191:

CryptonightR_instruction_mov192:

CryptonightR_instruction_mov193:

CryptonightR_instruction_mov194:

CryptonightR_instruction_mov195:

CryptonightR_instruction_mov196:

CryptonightR_instruction_mov197:
	mov	rcx, rax
CryptonightR_instruction_mov198:
	mov	rcx, rax
CryptonightR_instruction_mov199:

CryptonightR_instruction_mov200:

CryptonightR_instruction_mov201:

CryptonightR_instruction_mov202:

CryptonightR_instruction_mov203:

CryptonightR_instruction_mov204:

CryptonightR_instruction_mov205:
	mov	rcx, rax
CryptonightR_instruction_mov206:
	mov	rcx, rax
CryptonightR_instruction_mov207:

CryptonightR_instruction_mov208:

CryptonightR_instruction_mov209:

CryptonightR_instruction_mov210:

CryptonightR_instruction_mov211:

CryptonightR_instruction_mov212:

CryptonightR_instruction_mov213:
	mov	rcx, rax
CryptonightR_instruction_mov214:
	mov	rcx, rax
CryptonightR_instruction_mov215:

CryptonightR_instruction_mov216:

CryptonightR_instruction_mov217:

CryptonightR_instruction_mov218:

CryptonightR_instruction_mov219:

CryptonightR_instruction_mov220:

CryptonightR_instruction_mov221:
	mov	rcx, rax
CryptonightR_instruction_mov222:
	mov	rcx, rax
CryptonightR_instruction_mov223:

CryptonightR_instruction_mov224:

CryptonightR_instruction_mov225:

CryptonightR_instruction_mov226:

CryptonightR_instruction_mov227:

CryptonightR_instruction_mov228:

CryptonightR_instruction_mov229:
	mov	rcx, rdx
CryptonightR_instruction_mov230:
	mov	rcx, rdx
CryptonightR_instruction_mov231:

CryptonightR_instruction_mov232:

CryptonightR_instruction_mov233:

CryptonightR_instruction_mov234:

CryptonightR_instruction_mov235:

CryptonightR_instruction_mov236:

CryptonightR_instruction_mov237:
	mov	rcx, rdx
CryptonightR_instruction_mov238:
	mov	rcx, rdx
CryptonightR_instruction_mov239:

CryptonightR_instruction_mov240:

CryptonightR_instruction_mov241:

CryptonightR_instruction_mov242:

CryptonightR_instruction_mov243:

CryptonightR_instruction_mov244:

CryptonightR_instruction_mov245:
	mov	rcx, rdx
CryptonightR_instruction_mov246:
	mov	rcx, rdx
CryptonightR_instruction_mov247:

CryptonightR_instruction_mov248:

CryptonightR_instruction_mov249:

CryptonightR_instruction_mov250:

CryptonightR_instruction_mov251:

CryptonightR_instruction_mov252:

CryptonightR_instruction_mov253:
	mov	rcx, rdx
CryptonightR_instruction_mov254:
	mov	rcx, rdx
CryptonightR_instruction_mov255:

CryptonightR_instruction_mov256:

_TEXT_CN_TEMPLATE ENDS
END
